from pygeoinf.random_matrix import (
    fixed_rank_random_range,
    variable_rank_random_range,
    random_svd,
    random_eig,
    random_cholesky,
)

from pygeoinf.hilbert_space import (
    HilbertSpace,
    EuclideanSpace,
    Operator,
    LinearOperator,
    DiagonalLinearOperator,
    LinearForm,
)

from pygeoinf.gaussian_measure import (
    GaussianMeasure,
)

from pygeoinf.direct_sum import (
    HilbertSpaceDirectSum,
    BlockStructure,
    BlockLinearOperator,
    ColumnLinearOperator,
    RowLinearOperator,
    BlockDiagonalLinearOperator,
)

from pygeoinf.linear_solvers import (
    LinearSolver,
    DirectLinearSolver,
    LUSolver,
    CholeskySolver,
    IterativeLinearSolver,
    CGMatrixSolver,
    BICGMatrixSolver,
    BICGStabMatrixSolver,
    GMRESMatrixSolver,
    CGSolver,
)

from pygeoinf.forward_problem import ForwardProblem, LinearForwardProblem

from pygeoinf.optimisation import (
    LinearLeastSquaresInversion,
    LinearMinimumNormInversion,
)

from pygeoinf.bayesian import LinearBayesianInversion, LinearBayesianInference

from pygeoinf.checks.hilbert_space import HilbertSpaceChecks



"""
Module defined the forward problem class. 
"""

from scipy.stats import chi2
from pygeoinf.gaussian_measure import GaussianMeasure
from pygeoinf.direct_sum import BlockLinearOperator


class ForwardProblem:
    """
    Class for forward problems. A class instance is defined by
    setting the forward operator and the data error measure.
    """

    def __init__(self, forward_operator, data_error_measure=None):
        """
        Args:
            forward_operator (LinearOperator): Mapping from the model to data space.
            data_error_measure (GaussianMeasure): Gaussian measure from which data errors
                are assumed to be drawn. Default is None, in which case the data is taken
                to be error free.
        """
        self._forward_operator = forward_operator
        self._data_error_measure = data_error_measure
        if self.data_error_measure_set:
            assert self.data_space == data_error_measure.domain

    @property
    def forward_operator(self):
        """The forward operator."""
        return self._forward_operator

    @property
    def data_error_measure_set(self):
        """
        Returns true is data error measure has been set.
        """
        return self._data_error_measure is not None

    @property
    def data_error_measure(self):
        """The measure from which data errors are drawn."""
        if not self.data_error_measure_set:
            raise NotImplementedError("Data error measure has not been set")
        return self._data_error_measure

    @property
    def model_space(self):
        """The model space."""
        return self.forward_operator.domain

    @property
    def data_space(self):
        """The data space."""
        return self.forward_operator.codomain


class LinearForwardProblem(ForwardProblem):
    """
    Class for linear forward problems. A class instance is defined by
    setting the forward operator and the data error measure.

    Data error measures should be set either for not or for all of the
    forward problems.
    """

    @staticmethod
    def from_direct_sum(forward_problems):
        """
        Given a list of forward problems with a common model space, forms
        the joint forward problem.
        """

        forward_operator = BlockLinearOperator(
            [[forward_problem.forward_operator] for forward_problem in forward_problems]
        )

        if all(
            forward_problem.data_error_measure_set
            for forward_problem in forward_problems
        ):
            data_error_measure = GaussianMeasure.from_direct_sum(
                [
                    forward_problem.data_error_measure
                    for forward_problem in forward_problems
                ]
            )
        else:
            data_error_measure = None

        return LinearForwardProblem(
            forward_operator @ forward_operator.domain.subspace_inclusion(0),
            data_error_measure,
        )

    def data_measure(self, model):
        """Returns the data measure for a given model."""
        if not self.data_error_measure_set:
            raise NotImplementedError("Data error measure has not been set")
        return self.data_error_measure.affine_mapping(
            translation=self.forward_operator(model)
        )

    def synthetic_data(self, model):
        """Returns synthetic data corresponding to a given model."""
        return self.data_measure(model).sample()

    def synthetic_model_and_data(self, mu):
        """
        Given a Gaussian measure on the model space, returns a random model and
        the corresponding synthetic data.
        """
        u = mu.sample()
        if self.data_error_measure_set:
            return u, self.data_measure(u).sample()
        else:
            return u, self.forward_operator(u)

    def crtical_chi_squared(self, significance_level):
        """
        Returns the crtical value of chi-squared for a given significance level.
        """
        return chi2.ppf(significance_level, self.data_space.dim)

    def chi_squared(self, model, data):
        """
        Returns the chi-squared statistic for a given model and observed data. The a
        data error measure has not been set, this returns the squared norm of the
        data residual.

        Raises:
            NotImplementedError -- If the inverse covariance for the data error measure
                has not been set.

        """
        if self.data_error_measure_set:

            if self.data_error_measure.inverse_covariance is None:
                raise NotImplementedError("Inverse covariance has not been implemented")

            difference = self.data_space.subtract(data, self.forward_operator(model))
            difference = self.data_space.subtract(
                difference, self.data_error_measure.expectation
            )
            inverse_data_covariance = self.data_error_measure.inverse_covariance
            return self.data_space.inner_product(
                inverse_data_covariance(difference), difference
            )

        else:

            difference = self.data_space.subtract(data, self.forward_operator(model))
            return self.data_space.squared_norm(difference)

    def chi_squared_test(self, significance_level, model, data):
        """
        Returns True if the model is compatible the given data at the
        specified significance level.
        """
        return self.chi_squared(model, data) < self.crtical_chi_squared(
            significance_level
        )



"""
Module containing the base class for Sobolev spaces defined on homogeneous spaces. 
"""

from abc import ABC, abstractmethod
import numpy as np
from pygeoinf.hilbert_space import HilbertSpace, LinearOperator, EuclideanSpace


class HomogeneousSpaceSobolev(HilbertSpace, ABC):
    """
    Base class for Sobolev spaces of scalar fields on a homogeneous space.
    """

    def __init__(
        self,
        order,
        scale,
        dim,
        to_components,
        from_components,
        inner_product,
        to_dual,
        from_dual,
        /,
        *,
        add=None,
        subtract=None,
        multiply=None,
        axpy=None,
        copy=None,
        vector_multiply=None,
    ):
        """
        Args:
            order (float): Order of the Sobolev space.
            scale (float): Scale for the Sobolev space.
            dim (int): The dimension of the space, or of the
                finite-dimensional approximating space.
            to_components (callable):  A functor that maps vectors
                to their components.
            from_components (callable): A functor that maps components
                to vectors.
            inner_product (callable): A functor the implements the inner
                product on the space.
            to_dual (callable | None): A funcator that maps a vector
                to the cannonically associated dual vector.
            from_dual (callable | None): A functor that maps a dual vector
                to its representation on the space.
            add (callable): Implements vector addition.
            subtract (callable): Implements vector subtraction.
            multiply (callable): Implements scalar multiplication.
            axpy (callable): Implements the mapping y -> a*x + y
            copy (callable): Implements deep copy of a vector.
        """

        self._order = order

        if scale <= 0:
            raise ValueError("Scale must be positive")
        self._scale = scale

        super().__init__(
            dim,
            to_components,
            from_components,
            inner_product,
            to_dual,
            from_dual,
            add=add,
            subtract=subtract,
            multiply=multiply,
            axpy=axpy,
            copy=copy,
            vector_multiply=vector_multiply,
        )

    @property
    def order(self):
        """
        Return the Sobolev order.
        """
        return self._order

    @property
    def scale(self):
        """
        Return the Sobolev scale.
        """
        return self._scale

    @abstractmethod
    def random_point(self):
        """
        Returns a random point within the homogeneous space.
        """

    def random_points(self, n):
        """
        Returns a list of n random points.
        """
        return [self.random_point() for _ in range(n)]

    @abstractmethod
    def dirac(self, point):
        """
        Returns the diract measure based at the given point.
        """

    def dirac_representation(self, point):
        """
        Returns the representation of the Dirac measure based
        at the given point.
        """
        return self.from_dual(self.dirac(point))

    def point_evaluation_operator(self, points):
        """
        Returns as a linear operator the mapping from a function to
        its values at the list of points.
        """
        dim = len(points)
        matrix = np.zeros((dim, self.dim))

        for i, point in enumerate(points):
            cp = self.dirac(point).components
            matrix[i, :] = cp

        return LinearOperator.from_matrix(
            self, EuclideanSpace(dim), matrix, galerkin=True
        )

    @abstractmethod
    def invariant_automorphism(self, f):
        """
        Returns an automorphism of the form f(Delta) with
        Delta the Laplacian.
        """

    @abstractmethod
    def invariant_gaussian_measure(self, f, /, *, expectation=None):
        """
        Returns a Gaussian measure whose covariance takes the form
        f(Delta) with Delta the Laplaceian. An expectation can be
        provided, with the default being zero.
        """

    def _transform_measure(self, amplitude, expectation, mu):
        # Scales an invariant measure to get the right pointwise
        # variance and then adds the expected value.
        Q = mu.covariance
        u = self.dirac_representation(self.random_point())
        var = self.inner_product(Q(u), u)
        mu *= amplitude / np.sqrt(var)
        return mu.affine_mapping(translation=expectation)

    def sobolev_gaussian_measure(self, order, scale, amplitude, /, *, expectation=None):
        """
        Returns an invariant Gaussian measure whose covariance takes the Sobolev form.

        Args:
            order (float): Order parameter. Must be > 1.
            scale (float): Scale paramerer. Must be > 0.
            amplitude (float): Pointwise standard deviation. Must be > 0.
            expectation (vector): Expectation. Default is None.
        """
        mu = self.invariant_gaussian_measure(lambda k: (1 + scale**2 * k) ** -order)
        return self._transform_measure(amplitude, expectation, mu)

    def heat_gaussian_measure(self, scale, amplitude, /, *, expectation=None):
        """
        Returns an invariant Gaussian measure whose covariance takes the heat kernel form.

        Args:
            scale (float): Scale paramerer. Must be > 0.
            amplitude (float): Pointwise standard deviation. Must be > 0.
            expectation (vector): Expectation. Default is None.
        """
        mu = self.invariant_gaussian_measure(lambda k: np.exp(-(scale**2) * k))
        return self._transform_measure(amplitude, expectation, mu)



"""
Module containing the base class for inversion methods
"""


class Inversion:
    """
    Base class for inversion methods.
    """

    def __init__(self, forward_problem):
        self._forward_problem = forward_problem

    @property
    def forward_problem(self):
        """
        Return the forward problem.
        """
        return self._forward_problem

    @property
    def model_space(self):
        """
        Return the model space.
        """
        return self.forward_problem.model_space

    @property
    def data_space(self):
        """
        Return the data space.
        """
        return self.forward_problem.data_space

    @property
    def model_space_dim(self):
        """
        Return model space dimension.
        """
        return self.model_space.dim

    @property
    def data_space_dim(self):
        """
        Return data space dimension.
        """
        return self.data_space.dim

    def assert_data_error_measure(self):
        """
        Raises a not implemented error if a data error
        measure is not present
        """
        if not self.forward_problem.data_error_measure_set:
            raise NotImplementedError(
                "Data error measure requried for the inversion method"
            )

    def assert_inverse_data_covariance(self):
        """
        Raises a not implemented error if a data error measure does not have an
        inverse covariance set.
        """
        if (
            self.forward_problem.data_error_measure_set
            and not self.forward_problem.data_error_measure.inverse_covariance_set
        ):
            raise NotImplementedError(
                "Inverse data covariance needed for the inversion method"
            )



"""
 Sobolev spaces for functions on a circle. 
"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.fft import rfft, irfft
from scipy.sparse import diags
from pygeoinf.hilbert_space import (
    LinearOperator,
    LinearForm,
    EuclideanSpace,
)
from pygeoinf.gaussian_measure import GaussianMeasure
from pygeoinf.homogeneous_space.homogeneous_space import HomogeneousSpaceSobolev


class Sobolev(HomogeneousSpaceSobolev):
    """
    Implementation of the Sobolev space H^s on a circle.
    """

    def __init__(
        self,
        kmax,
        order,
        scale,
        /,
        *,
        radius=1,
    ):
        """
        Args:
            kmax (float): The maximum Fourier degree.
            order (float): Sobolev order.
            scale (float): Sobolev length-scale.
            radius (float): Radius of the circle. Default is 1.
        """

        self._kmax = kmax
        self._radius = radius

        super().__init__(
            order,
            scale,
            2 * kmax,
            self._to_componets,
            self._from_componets,
            self._inner_product,
            self._to_dual,
            self._from_dual,
            vector_multiply=lambda u1, u2: u1 * u2,
        )

        self._fft_factor = np.sqrt(2 * np.pi * radius) / self.dim
        self._inverse_fft_factor = 1 / self._fft_factor

        values = np.zeros(self.kmax + 1)
        values[0] = 1
        for k in range(1, self.kmax + 1):
            values[k] = 2 * self._sobolev_function(k)

        self._metric = diags([values], [0])
        self._inverse_metric = diags([np.reciprocal(values)], [0])

    @staticmethod
    def from_sobolev_parameters(
        order, scale, /, *, radius=1, rtol=1e-8, power_of_two=False
    ):
        """
        Returns a instance of the class with the maximum Fourier degree
        chosen based on the Sobolev parameters. The method is based on
        an estimate of truncation error for the Dirac measure, and is
        only applicable in spaces with orders > 0.5.

        Args:
            order (float): The Sobolev order.
            scale (float): The Sobolev length-scale.
            radius (float): The radius for the circle. Default is 1.
            rtol (float): Relative tolerance used in assessing
                truncation error. Default is 1e-8
            power_of_two (bool): If True, truncation degree set to optimal
                power of two.

        Raises:
            ValueError: If order is <= 0.5.
        """

        if order <= 0.5:
            raise ValueError("This method is only applicable for orders > 0.5")

        summation = 1
        k = 0
        err = 1
        while err > rtol:
            k += 1
            term = (1 + (scale * k / radius) ** 2) ** -order
            summation += term
            err = term / summation

        if power_of_two:
            n = int(np.log2(k))
            k = 2 ** (n + 1)

        return Sobolev(k, order, scale, radius=radius)

    @property
    def kmax(self):
        """
        Return the maximum Fourier degree.
        """
        return self._kmax

    @property
    def radius(self):
        """
        Return the radius.
        """
        return self._radius

    @property
    def angle_spacing(self):
        """
        Return the angle spacing.
        """
        return 2 * np.pi / self.dim

    def random_point(self):
        return np.random.uniform(0, 2 * np.pi)

    def angles(self):
        """
        Returns a numpy array of the angles.
        """
        return np.fromiter(
            [i * self.angle_spacing for i in range(self.dim)],
            float,
        )

    def project_function(self, f):
        """
        Returns an element of the space formed by projecting a given function.
        """
        return np.fromiter([f(theta) for theta in self.angles()], float)

    def plot(self, u, fig=None, ax=None, **kwargs):
        """
        Make a simple plot of an element of the space on the computational domain.

        Args:
            u (vector): The element of the space to be plotted.
            fig (Figure): An existing Figure object to use. Default is None.
            ax (Axes): An existing Axes object to use. Default is None.
            kwargs: Keyword arguments forwarded to plot.

        Returns
            Figure: The figure object, either that given or newly created.
            Axes: The axes object, either that given or newly created.
        """

        figsize = kwargs.pop("figsize", (10, 8))

        if fig is None:
            fig = plt.figure(figsize=figsize)
        if ax is None:
            ax = fig.add_subplot()

        ax.plot(self.angles(), u, **kwargs)

        return fig, ax

    def plot_error_bounds(self, u, u_bound, fig=None, ax=None, **kwargs):
        """
        Make a plot of an element of the space bounded above and below by a standard
        deviation curve.

        Args:
            u (vector): The element of the space to be plotted.
            u_bounds (vector): A second element giving point-wise bounds.
            fig (Figure): An existing Figure object to use. Default is None.
            ax (Axes): An existing Axes object to use. Default is None.
            kwargs: Keyword arguments forwarded to plot.

        Returns
            Figure: The figure object, either that given or newly created.
            Axes: The axes object, either that given or newly created.
        """

        figsize = kwargs.pop("figsize", (10, 8))

        if fig is None:
            fig = plt.figure(figsize=figsize)
        if ax is None:
            ax = fig.add_subplot()

        ax.fill_between(self.angles(), u - u_bound, u + u_bound, **kwargs)

        return fig, ax

    def invariant_automorphism(self, f):
        values = np.fromiter(
            [f(k * k / self.radius**2) for k in range(self.kmax + 1)], dtype=float
        )
        matrix = diags([values], [0])

        def mapping(u):
            coeff = self.to_coefficient(u)
            coeff = matrix @ coeff
            return self.from_coefficient(coeff)

        return LinearOperator.formally_self_adjoint(self, mapping)

    def invariant_gaussian_measure(self, f, /, *, expectation=None):
        values = np.fromiter(
            [np.sqrt(f(k * k / self.radius**2)) for k in range(self.kmax + 1)],
            dtype=float,
        )
        matrix = diags([values], [0])

        domain = EuclideanSpace(self.dim)
        codomain = self

        def mapping(c):
            coeff = self._component_to_coefficient(c)
            coeff = matrix @ coeff
            return self.from_coefficient(coeff)

        def formal_adjoint(u):
            coeff = self.to_coefficient(u)
            coeff = matrix @ coeff
            return self._coefficient_to_component(coeff)

        covariance_factor = LinearOperator.from_formal_adjoint(
            domain, codomain, mapping, formal_adjoint
        )

        return GaussianMeasure(
            covariance_factor=covariance_factor,
            expectation=expectation,
        )

    def dirac(self, point):
        coeff = np.zeros(self.kmax + 1, dtype=complex)
        fac = np.exp(-1j * point)
        coeff[0] = 1
        for k in range(1, coeff.size):
            coeff[k] = coeff[k - 1] * fac
        coeff *= 1 / np.sqrt(2 * np.pi * self.radius)
        coeff[1:] *= 2
        cp = self._coefficient_to_component(coeff)
        return LinearForm(self, components=cp)

    def to_coefficient(self, u):
        """
        Maps an element to its Fourier coefficients.
        """
        return rfft(u) * self._fft_factor

    def from_coefficient(self, coeff):
        """
        Maps Fourier coefficients to an element.
        """
        return irfft(coeff, n=self.dim) * self._inverse_fft_factor

    # ================================================================#
    #                         Private methods                         #
    # ================================================================#

    def _sobolev_function(self, k):
        return (1 + (self.scale * k) ** 2) ** self.order

    def _coefficient_to_component(self, coeff):
        return np.concatenate([coeff.real, coeff.imag[1 : self.kmax]])

    def _component_to_coefficient(self, c):
        coeff_real = c[: self.kmax + 1]
        coeff_imag = np.concatenate([[0], c[self.kmax + 1 :], [0]])
        return coeff_real + 1j * coeff_imag

    def _to_componets(self, u):
        coeff = self.to_coefficient(u)
        return self._coefficient_to_component(coeff)

    def _from_componets(self, c):
        coeff = self._component_to_coefficient(c)
        return self.from_coefficient(coeff)

    def _inner_product(self, u1, u2):
        coeff1 = self.to_coefficient(u1)
        coeff2 = self.to_coefficient(u2)
        return np.real(np.vdot(self._metric @ coeff1, coeff2))

    def _to_dual(self, u):
        coeff = self.to_coefficient(u)
        cp = self._coefficient_to_component(self._metric @ coeff)
        return self.dual.from_components(cp)

    def _from_dual(self, up):
        cp = self.dual.to_components(up)
        coeff = self._component_to_coefficient(cp)
        c = self._coefficient_to_component(self._inverse_metric @ coeff)
        return self.from_components(c)


class Lebesgue(Sobolev):
    """
    Implementation of the Lebesgue space L2 on a circle.
    """

    def __init__(
        self,
        kmax,
        /,
        *,
        radius=1,
    ):
        """
        Args:
            kmax (float): The maximum Fourier degree.
            radius (float): Radius of the circle. Default is 1.
        """
        super().__init__(kmax, 0, 1, radius=radius)
